from dataclasses import dataclass
from enum import Enum
from typing import Callable
from typing import Optional

from dataclasses_json import dataclass_json


class PKaAttribute:
    ACID_1 = "pka_acid1"
    BASE_1 = "pka_base1"
    ACID_2 = "pka_acid2"
    BASE_2 = "pka_base2"


class AcidType(Enum):
    STRONG = "strong_acid"
    WEAK = "weak_acid"
    NONE = "no_acid"


class BaseType(Enum):
    STRONG = "strong_base"
    WEAK = "weak_base"
    NONE = "no_base"


@dataclass_json
@dataclass
class ABClassData:
    """Class for tracking the types of Acid/Base groups in a molecule."""

    acid_1_class: Optional[Enum] = None
    acid_2_class: Optional[Enum] = None
    base_1_class: Optional[Enum] = None
    base_2_class: Optional[Enum] = None

    # references to own field names
    ACID_1_FIELD = "acid_1_class"
    ACID_2_FIELD = "acid_2_class"
    BASE_1_FIELD = "base_1_class"
    BASE_2_FIELD = "base_2_class"

    def convert_nones_to_nulls(self) -> None:
        """Utility method to convert AcidType.NONE and BaseType.NONE to None."""
        for k in self.__dict__.keys():
            if getattr(self, k) in (AcidType.NONE, BaseType.NONE):
                setattr(self, k, None)


@dataclass
class Rule:
    """Rule to determine a class."""

    operator: Callable
    type: Enum
    value: Optional[float]



import pytest

from abcount import ABClassBuilder
from abcount import PKaClassBuilder
from abcount.components.classifier import ABClassData
from abcount.components.classifier import AcidType
from abcount.components.classifier import BaseType


@pytest.mark.parametrize(
    "input_dict,expected",
    [
        (
            {
                "pka_acid1": 3.5,
                "pka_acid2": None,
                "pka_base1": 3.785,
                "pka_base2": None,
            },
            ABClassData(
                acid_1_class=AcidType.STRONG,
                acid_2_class=AcidType.NONE,
                base_1_class=BaseType.NONE,
                base_2_class=BaseType.NONE,
            ),
        ),
        (
            {"pka_acid1": 3.5, "pka_acid2": 4.5, "pka_base1": 3.785, "pka_base2": None},
            ABClassData(
                acid_1_class=AcidType.STRONG,
                acid_2_class=AcidType.STRONG,
                base_1_class=BaseType.NONE,
                base_2_class=BaseType.NONE,
            ),
        ),
        (
            {"pka_acid1": 6.3, "pka_acid2": 7.3, "pka_base1": 3.785, "pka_base2": None},
            ABClassData(
                acid_1_class=AcidType.STRONG,
                acid_2_class=AcidType.WEAK,
                base_1_class=BaseType.NONE,
                base_2_class=BaseType.NONE,
            ),
        ),
        (
            {"pka_acid1": 6.4, "pka_acid2": 7.4, "pka_base1": 3.785, "pka_base2": None},
            ABClassData(
                acid_1_class=AcidType.WEAK,
                acid_2_class=AcidType.NONE,
                base_1_class=BaseType.NONE,
                base_2_class=BaseType.NONE,
            ),
        ),
        (
            {
                "pka_acid1": None,
                "pka_acid2": 7.3,
                "pka_base1": 7.785,
                "pka_base2": None,
            },
            ABClassData(
                acid_1_class=AcidType.NONE,
                acid_2_class=AcidType.WEAK,
                base_1_class=BaseType.WEAK,
                base_2_class=BaseType.NONE,
            ),
        ),
        (
            {
                "pka_acid1": None,
                "pka_acid2": 7.3,
                "pka_base1": 9.785,
                "pka_base2": None,
            },
            ABClassData(
                acid_1_class=AcidType.NONE,
                acid_2_class=AcidType.WEAK,
                base_1_class=BaseType.STRONG,
                base_2_class=BaseType.NONE,
            ),
        ),
        (
            {"pka_acid1": None, "pka_acid2": 7.3, "pka_base1": 9.785, "pka_base2": 7.7},
            ABClassData(
                acid_1_class=AcidType.NONE,
                acid_2_class=AcidType.WEAK,
                base_1_class=BaseType.STRONG,
                base_2_class=BaseType.WEAK,
            ),
        ),
        (
            {"pka_acid1": None, "pka_acid2": 7.3, "pka_base1": 9.785, "pka_base2": 7.7},
            ABClassData(
                acid_1_class=AcidType.NONE,
                acid_2_class=AcidType.WEAK,
                base_1_class=BaseType.STRONG,
                base_2_class=BaseType.WEAK,
            ),
        ),
        (
            {
                "pka_acid1": None,
                "pka_acid2": None,
                "pka_base1": 7.785,
                "pka_base2": 7.7,
            },
            ABClassData(
                acid_1_class=AcidType.NONE,
                acid_2_class=AcidType.NONE,
                base_1_class=BaseType.WEAK,
                base_2_class=BaseType.WEAK,
            ),
        ),
    ],
)
def test_abclass_prediction(input_dict, expected):
    classes_obj = ABClassBuilder().build(input_dict)
    assert classes_obj == expected


@pytest.mark.parametrize(
    "input_dict,expected",
    [
        (
            {"my_pkaa1": 3.5, "pka_acid2": None, "my_pkab_1": 3.785, "pka_base2": None},
            ABClassData(
                acid_1_class=AcidType.STRONG,
                acid_2_class=AcidType.NONE,
                base_1_class=BaseType.NONE,
                base_2_class=BaseType.NONE,
            ),
        )
    ],
)
def test_abclass_custom_builder(input_dict, expected):
    CustomPKaAttribute = PKaClassBuilder.build(ACID_1="my_pkaa1", BASE_1="my_pkab_1")
    classes_obj = ABClassBuilder().build(
        input_dict, pka_attribute_cls=CustomPKaAttribute
    )
    assert classes_obj == expected


@pytest.mark.parametrize(
    "input_dict,expected",
    [
        (
            {"my_pkaa1": 3.5, "my_pkab_1": 9.785},
            ABClassData(acid_1_class=AcidType.STRONG, base_1_class=BaseType.STRONG),
        )
    ],
)
def test_abclass_custom_builder_and_groups(input_dict, expected):
    CustomPKaAttribute = PKaClassBuilder.build(ACID_1="my_pkaa1", BASE_1="my_pkab_1")
    classes_obj = ABClassBuilder().build(
        input_dict, pka_attribute_cls=CustomPKaAttribute, num_acids=1, num_bases=1
    )
    assert classes_obj == expected


@pytest.mark.parametrize(
    "input_dict,expected",
    [
        (
            {"my_pkaa_1": 3.5, "my_pkaa_2": 7.5, "my_pkab_1": 9.785},
            ABClassData(acid_1_class=AcidType.STRONG, base_1_class=BaseType.STRONG),
        )
    ],
)
def test_abclass_json(input_dict, expected):
    CustomPKaAttribute = PKaClassBuilder.build(
        ACID_1="my_pkaa_1", ACID_2="my_pkaa_2", BASE_1="my_pkab_1"
    )
    classes_obj = ABClassBuilder().build(
        input_dict, pka_attribute_cls=CustomPKaAttribute, num_acids=2, num_bases=1
    )
    expected = '{"acid_1_class": "strong_acid", "acid_2_class": "no_acid", "base_1_class": "strong_base", "base_2_class": null}'  # noqa
    assert classes_obj.to_json() == expected



from enum import Enum

from abcount.components.classifier import ABClassData
from abcount.model.classifier import AcidType
from abcount.model.classifier import BaseType
from abcount.model.ions import IonDefinition


class IonMatcher:
    """Matches AB data to the ion and species definitions."""

    def match_class_data(self, class_data: ABClassData) -> IonDefinition:
        # ignore difference AcidType.NONE | BaseType.NONE and None
        class_data.convert_nones_to_nulls()
        for ion_definition in IonRules:
            ion_obj = ion_definition.value
            if class_data == ion_obj.class_data:
                return ion_obj
        # raise ValueError(f"Could not match any ion definition - (got {class_data})")
        return IonDefinition(
            class_data=class_data,
            major_species_ph74_class="undefined",
            ion_class="undefined",
            explanation="",
        )


class IonRules(Enum):
    I1 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="zwitterion",
        explanation="zwitterion",
    )

    I2 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="zwitterion",
        explanation="zwitterion - weak base",
    )

    I3 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=AcidType.STRONG,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="dianion",
        ion_class="diacid",
        explanation="diacid",
    )

    I4 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=AcidType.WEAK,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="dianion",
        ion_class="diacid",
        explanation="diacid - strong + weak",
    )

    I5 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=None,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="anion",
        ion_class="acid",
        explanation="acid",
    )

    I6 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=None,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="anion",
        ion_class="weak acid",
        explanation="weak acid",
    )

    I7 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=AcidType.WEAK,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="dianion",
        ion_class="weak diacid",
        explanation="weak diacid",
    )

    I8 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=None,
        ),
        major_species_ph74_class="cation",
        ion_class="base",
        explanation="base",
    )

    I9 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=BaseType.STRONG,
        ),
        major_species_ph74_class="dication",
        ion_class="dibase",
        explanation="dibase",
    )

    I10 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=None,
        ),
        major_species_ph74_class="cation",
        ion_class="weak base",
        explanation="weak base",
    )

    I11 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=BaseType.WEAK,
        ),
        major_species_ph74_class="dication",
        ion_class="weak dibase",
        explanation="weak dibase",
    )

    I12 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=BaseType.WEAK,
        ),
        major_species_ph74_class="dication",
        ion_class="dibase",
        explanation="dibase - strong + weak",
    )

    I13 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="weak zwitterion",
        explanation="weak zwitterion",
    )

    I14 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="zwitterion",
        explanation="zwitterion - weak acid",
    )

    I15 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None, acid_2_class=None, base_1_class=None, base_2_class=None
        ),
        major_species_ph74_class="neutral",
        ion_class="neutral",
        explanation="neutral",
    )



import os
from dataclasses import dataclass
from pathlib import Path


@dataclass
class ProjectDirs:
    package_dir: Path
    source_dir: Path
    data_dir: Path


_source_dir = os.path.dirname(__file__)
_package_dir = os.path.dirname(_source_dir)
_data_dir = os.path.join(_source_dir, "data")
dirs = ProjectDirs(
    package_dir=Path(_package_dir),
    source_dir=Path(_source_dir),
    data_dir=Path(_data_dir),
)


@dataclass
class DataFiles:
    raw_data_file: Path
    acid_defs_file: Path
    base_defs_file: Path


_files = DataFiles(
    raw_data_file=Path("raw_definitions.json"),
    acid_defs_file=Path("acid_definitions.json"),
    base_defs_file=Path("base_definitions.json"),
)


@dataclass
class DataFilePaths:
    raw_data_filepath: Path
    acid_defs_filepath: Path
    base_defs_filepath: Path


fps = DataFilePaths(
    raw_data_filepath=Path(os.path.join(dirs.data_dir, _files.raw_data_file)),
    acid_defs_filepath=Path(os.path.join(dirs.data_dir, _files.acid_defs_file)),
    base_defs_filepath=Path(os.path.join(dirs.data_dir, _files.base_defs_file)),
)



import numpy as np

from abcount.model.classifier import ABClassData
from abcount.model.classifier import AcidType
from abcount.model.classifier import BaseType
from abcount.model.classifier import PKaAttribute
from abcount.model.classifier import Rule


def isnone(value, _):
    return value is None


class AcidRule:
    NONE = Rule(operator=isnone, type=AcidType.NONE, value=None)
    STRONG = Rule(operator=np.less, type=AcidType.STRONG, value=6.4)
    WEAK = Rule(operator=np.less, type=AcidType.WEAK, value=7.4)
    NEGLEGIBLE = Rule(operator=np.greater_equal, type=AcidType.NONE, value=7.4)


class BaseRule:
    NONE = Rule(operator=isnone, type=BaseType.NONE, value=None)
    STRONG = Rule(operator=np.greater, type=BaseType.STRONG, value=8.4)
    WEAK = Rule(operator=np.greater, type=BaseType.WEAK, value=7.4)
    NEGLEGIBLE = Rule(operator=np.less_equal, type=BaseType.NONE, value=7.4)


class ABClassBuilder:
    """Maps pKa data to acid/base classes according to
    a customisable set of rules."""

    def build(
        self, input: dict, pka_attribute_cls=PKaAttribute, num_acids=2, num_bases=2
    ):
        self.classes_obj = ABClassData()
        self.__pka_attribute_cls__ = pka_attribute_cls
        self.num_acids = num_acids
        self.num_bases = num_bases
        self._set_attribute_lists()
        self._build_attribute_map()
        self.predict_input(input)
        return self.classes_obj

    def _build_attribute_map(self) -> None:
        self.__pka_to_group_map__ = {
            self.__pka_attribute_cls__.ACID_1: ABClassData.ACID_1_FIELD,
            self.__pka_attribute_cls__.ACID_2: ABClassData.ACID_2_FIELD,
            self.__pka_attribute_cls__.BASE_1: ABClassData.BASE_1_FIELD,
            self.__pka_attribute_cls__.BASE_2: ABClassData.BASE_2_FIELD,
        }

    def _set_attribute_lists(self) -> None:
        self.acid_attribute_list = [
            self.__pka_attribute_cls__.ACID_1,
            self.__pka_attribute_cls__.ACID_2,
        ][: self.num_acids]
        self.base_attribute_list = [
            self.__pka_attribute_cls__.BASE_1,
            self.__pka_attribute_cls__.BASE_2,
        ][: self.num_bases]

    def predict_input(self, input: dict) -> str:
        self._classify_acids(input)
        self._classify_bases(input)

    def _classify_acids(self, input):
        self._classify_group(
            input,
            # ordered by priority of execution
            rules=[AcidRule.NONE, AcidRule.STRONG, AcidRule.WEAK, AcidRule.NEGLEGIBLE],
            attributes=self.acid_attribute_list,
        )

    def _classify_bases(self, input):
        self._classify_group(
            input,
            # ordered by priority of execution
            rules=[BaseRule.NONE, BaseRule.STRONG, BaseRule.WEAK, BaseRule.NEGLEGIBLE],
            attributes=self.base_attribute_list,
        )

    def _classify_group(self, input, rules, attributes) -> None:
        """Match pKa values to classes."""
        for pka_attr in attributes:
            pka_value = input[pka_attr]
            for r in rules:
                func = r.operator
                ref_value = r.value
                ion_attr = self.__pka_to_group_map__[pka_attr]
                if func(pka_value, ref_value):
                    assigned_class = r.type
                    setattr(self.classes_obj, ion_attr, assigned_class)
                    break


class PKaClassBuilder:
    @classmethod
    def build(cls, **overrides):
        """
        Return a CustomPKaAttribute class with overridden values.
        Any attributes not given in overrides will
        fall back to the parent values.
        """
        return type("CustomPKaAttribute", (PKaAttribute,), overrides)


if __name__ == "__main__":
    example = {
        "pka_acid1": 3.5,
        "pka_acid2": None,
        "pka_base1": 3.785,
        "pka_base2": None,
    }
    print(example)
    mjc = ABClassBuilder()
    mjc.predict_from_dict(example)
    mjc.get_classes_obj()



from dataclasses import dataclass

from dataclasses_json import dataclass_json

from abcount.model.classifier import ABClassData


@dataclass_json
@dataclass
class IonDefinition:
    """Class for storing class conditions and ionic definitions."""

    class_data: ABClassData
    major_species_ph74_class: str
    ion_class: str
    explanation: str



import json
from pathlib import Path

from rdkit import Chem

from abcount.config import fps
from abcount.model.counter import DefinitionAttribute
from abcount.model.counter import GroupTypeAttribute


class RawDefinitionAttribute:
    SMARTS = "SMARTS"
    TYPE = "type"
    ACIDIC_TYPE = "A"
    BASIC_TYPE = "B"


def _validate_smarts(smarts: str) -> None:
    """Raise an error if the SMARTS pattern is invalid."""
    Chem.MolFromSmarts(smarts)


def preprocess_definition(def_dict: dict) -> dict:
    """
    Normalize and validate a SMARTS definition.
    The function will fail loudly on malformed input.
    """
    smarts = def_dict[RawDefinitionAttribute.SMARTS]
    raw_type = def_dict[RawDefinitionAttribute.TYPE]

    if raw_type == RawDefinitionAttribute.ACIDIC_TYPE:
        def_type = GroupTypeAttribute.ACID
    elif raw_type == RawDefinitionAttribute.BASIC_TYPE:
        def_type = GroupTypeAttribute.BASE
    else:
        def_type = raw_type  # Already normalized

    _validate_smarts(smarts)

    return {DefinitionAttribute.SMARTS: smarts, DefinitionAttribute.TYPE: def_type}


class SmartsStore:
    """Stores and organizes SMARTS definitions by type."""

    def __init__(self):
        self._acid_defs = []
        self._base_defs = []

    def add(self, definition: dict) -> None:
        def_type = definition[DefinitionAttribute.TYPE]
        if def_type == GroupTypeAttribute.ACID:
            self._acid_defs.append(definition)
        elif def_type == GroupTypeAttribute.BASE:
            self._base_defs.append(definition)
        else:
            raise ValueError(f"Unknown group type: {def_type}")

    @property
    def acid_defs(self) -> list:
        return self._acid_defs

    @property
    def base_defs(self) -> list:
        return self._base_defs


class SmartsDump:
    """Utility for writing SMARTS definitions to disk."""

    @staticmethod
    def save_definitions_to_json(defs: list, filepath: Path) -> None:
        with open(filepath, "w") as f:
            json.dump(defs, f, indent=4)


if __name__ == "__main__":
    with open(fps.raw_data_filepath) as f:
        definitions = json.load(f)

    store = SmartsStore()

    for raw_def in definitions:
        processed_def = preprocess_definition(raw_def)
        store.add(processed_def)

    SmartsDump.save_definitions_to_json(store.acid_defs, fps.acid_defs_filepath)
    SmartsDump.save_definitions_to_json(store.base_defs, fps.base_defs_filepath)



from rdkit import Chem

from abcount import ABCounter
from abcount.components import SmartsMatcher


def test_smarts_matcher_1():
    matcher = SmartsMatcher(["[OX2H]"])
    example_mol = Chem.MolFromSmiles("C(O)COCO")
    res = matcher.count_matches(example_mol)
    expected = 2
    assert res == expected


def test_smarts_matcher_2():
    matcher = SmartsMatcher(["[n;r5][nH;r5][c;r5]"])
    example_mol = Chem.MolFromSmiles("[nH]1nnnc1-c3c2[nH]ncc2ccc3")
    res = matcher.count_matches(example_mol)
    expected = 2
    assert res == expected


def test_abcounter_1():
    smiles = "[nH]1nnnc1-c3c2[nH]ncc2ccc3"
    acid_exp = 2
    base_exp = 2
    mol = Chem.MolFromSmiles(smiles)
    abc = ABCounter()
    res = abc.count_acid_and_bases(mol)
    acid_pred = res["acid"]
    base_pred = res["base"]
    assert acid_exp == acid_pred
    assert base_exp == base_pred



import logging

import rdkit
from rdkit import Chem

from abcount.components.smarts import SmartsMatcherJson
from abcount.config import fps
from abcount.model.counter import GroupTypeAttribute


logger = logging.getLogger(__name__)


class ABCounter:
    def __init__(
        self,
        acid_defs_filepath=fps.acid_defs_filepath,
        base_defs_filepath=fps.base_defs_filepath,
    ):
        """
        Counts acidic and basic functional groups in a molecule using SMARTS pattern matching.

        Args:
            acid_defs_filepath (str): Path to custom acidic SMARTS definitions (optional).
            base_defs_filepath (str): Path to custom basic SMARTS definitions (optional).

        Attributes:
            acid_matcher (SmartsMatcherJson): Matcher for acidic functional group patterns.
            base_matcher (SmartsMatcherJson): Matcher for basic functional group patterns.
        """
        self.acid_defs_filepath = acid_defs_filepath
        self.base_defs_filepath = base_defs_filepath
        logger.debug(f"Loading Acidic SMARTS from path: {self.acid_defs_filepath}")
        logger.debug(f"Loading Basic SMARTS from path: {self.base_defs_filepath}")
        self.acid_matcher = SmartsMatcherJson(self.acid_defs_filepath)
        self.base_matcher = SmartsMatcherJson(self.base_defs_filepath)

    def count_acid_and_bases(self, mol: rdkit.Chem.Mol):
        """
        Count acidic and basic groups in the given molecule.

        Args:
            mol (rdkit.Chem.Mol): An RDKit molecule object.

        Returns:
            dict: A dictionary with counts keyed by GroupTypeAttribute.ACID and
                  GroupTypeAttribute.BASE.
        """
        acid_count = self.acid_matcher.count_matches(mol)
        base_count = self.base_matcher.count_matches(mol)
        return {
            GroupTypeAttribute.ACID: acid_count,
            GroupTypeAttribute.BASE: base_count,
        }


if __name__ == "__main__":
    # Test
    smiles = "OC(CC)(C)C#C"
    acid_exp = 0
    base_exp = 3
    mol = Chem.MolFromSmiles(smiles)
    abc = ABCounter()
    print(abc.count_acid_and_bases(mol))



import _match
import _report
import pytest

from abcount import ABCounter


abcounter = ABCounter()
validator = _match.ABValidator
entry_factory = _report.ReportEntryFactory


class SmartsMatcherMock:
    def generate_matches_list(self, mol):
        return [({"smarts": "foo", "type": "bar"}, 2)]


def test_tp_outcome_excess():
    outcome = validator.generate_outcome(expected_groups=2, predicted_groups=4)
    assert isinstance(outcome, _match.TruePositive)


def test_tp_outcome_defect():
    outcome = validator.generate_outcome(expected_groups=2, predicted_groups=1)
    assert isinstance(outcome, _match.TruePositive)


def test_fp_outcome():
    outcome = validator.generate_outcome(expected_groups=1, predicted_groups=2)
    assert isinstance(outcome, _match.FalsePositive)


def test_fn_outcome():
    outcome = validator.generate_outcome(expected_groups=1, predicted_groups=0)
    assert isinstance(outcome, _match.FalseNegative)


def test_fp_reporter():
    fp_outcome = validator.generate_outcome(expected_groups=1, predicted_groups=2)
    entry = entry_factory.generate_entry(
        SmartsMatcherMock(), fp_outcome, "FOO"
    )  # valid mock smiles
    reporter = _report.ReportGenerator()
    reporter.add(entry)
    expected = {
        "target": {0: "FOO"},
        "smarts": {0: "foo"},
        "matches": {0: 2},
        "total_expected_matches": {0: 1},
    }  # noqa
    assert reporter.fp_df.to_dict() == expected


def test_fp_runtime_error_reporter():
    with pytest.raises(RuntimeError):
        fp_outcome = validator.generate_outcome(expected_groups=1, predicted_groups=2)
        _ = entry_factory.generate_entry(
            SmartsMatcherMock(), fp_outcome, "FOOBAr"
        )  # invalid mock smiles


def test_fn_reporter():
    fn_outcome = validator.generate_outcome(expected_groups=1, predicted_groups=0)
    entry = entry_factory.generate_entry(
        SmartsMatcherMock(), fn_outcome, "FOO"
    )  # valid mock smiles
    reporter = _report.ReportGenerator()
    reporter.add(entry)
    expected = {"target": {0: "FOO"}, "total_expected_matches": {0: 1}}  # noqa
    assert reporter.fn_df.to_dict() == expected



from abcount import _logging  # noqa
from abcount.components.classifier import ABClassBuilder  # noqa
from abcount.components.classifier import PKaClassBuilder  # noqa
from abcount.components.counter import ABCounter  # noqa
from abcount.components.ions import IonMatcher  # noqa



from abcount import IonMatcher
from abcount.components.classifier import ABClassData
from abcount.model.classifier import AcidType
from abcount.model.classifier import BaseType


ion_matcher = IonMatcher()


def test_ions_matcher_1():
    abcd = ABClassData(
        acid_1_class=AcidType.STRONG,
        acid_2_class=None,
        base_1_class=BaseType.STRONG,
        base_2_class=None,
    )
    ion_def = ion_matcher.match_class_data(abcd)
    assert ion_def.major_species_ph74_class == "zwitterion"
    assert ion_def.ion_class == "zwitterion"
    assert ion_def.explanation == "zwitterion"


def test_ions_matcher_json_1():
    abcd = ABClassData(
        acid_1_class=AcidType.STRONG,
        acid_2_class=None,
        base_1_class=BaseType.STRONG,
        base_2_class=None,
    )
    ion_def = ion_matcher.match_class_data(abcd).to_json()
    expected = '{"class_data": {"acid_1_class": "strong_acid", "acid_2_class": null, "base_1_class": "strong_base", "base_2_class": null}, "major_species_ph74_class": "zwitterion", "ion_class": "zwitterion", "explanation": "zwitterion"}'  # noqa
    assert ion_def == expected


def test_ions_matcher_ignore_nones():
    abcd = ABClassData(
        acid_1_class=AcidType.STRONG,
        acid_2_class=AcidType.NONE,  # ignored
        base_1_class=BaseType.STRONG,
        base_2_class=None,
    )
    ion_def = ion_matcher.match_class_data(abcd)
    assert ion_def.major_species_ph74_class == "zwitterion"
    assert ion_def.ion_class == "zwitterion"
    assert ion_def.explanation == "zwitterion"


def test_ions_matcher_undefined():
    abcd = ABClassData(
        acid_1_class=AcidType.STRONG,
        acid_2_class=None,
        base_1_class=BaseType.STRONG,
        base_2_class=BaseType.STRONG,
    )
    ion_def = ion_matcher.match_class_data(abcd)
    assert ion_def.major_species_ph74_class == "undefined"
    assert ion_def.ion_class == "undefined"
    assert ion_def.explanation == ""



import json
import logging

from rdkit import Chem


logger = logging.getLogger(__name__)


class SmartsMatcher:
    """Plain SMARTS matcher with count functionality."""

    def __init__(self, definitions_list: list):
        self._load_smarts(definitions_list)

    def _load_smarts(self, definitions_list):
        # Loaded as a tuple to retain original SMARTS for debugging
        self.definitions_tup = [(Chem.MolFromSmarts(d), d) for d in definitions_list]

    def count_matches(self, mol):
        count = 0
        for t in self.definitions_tup:
            matches = SmartsMatcher._match_smarts_to_mol(mol, t[0])
            if matches:
                logger.debug(f"SMARTS: {t[1]} - Matches: {matches}")
            count += matches
        return count

    def generate_matches_list(self, mol):
        matches_list = []
        for t in self.definitions_tup:
            matches = SmartsMatcher._match_smarts_to_mol(mol, t[0])
            if matches:
                matches_list.append((t[1], matches))
        return matches_list

    @staticmethod
    def _match_smarts_to_mol(mol, smarts):
        return len(mol.GetSubstructMatches(smarts))


class SmartsMatcherJson(SmartsMatcher):
    """Adapter for SmartsMatcher."""

    def __init__(self, definitions_fp):
        self.definitions_fp = definitions_fp
        self._load_smarts()

    def _load_smarts(self):
        definitions_dict = self._load_json()
        # Loaded as a tuple to retain original SMARTS for debugging
        self.definitions_tup = [
            (Chem.MolFromSmarts(d["smarts"]), d) for d in definitions_dict
        ]

    def _load_json(self):
        with open(self.definitions_fp) as f:
            return json.load(f)
